#!/usr/bin/env python3

# Copyright (C) 2021  Jimmy Sie

# Version 2.0

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import xml.etree.ElementTree as ET
import os
import sys
import shutil
import dateutil.parser
import time
import filetype
import hashlib
from datetime import timezone

if getattr(sys, 'frozen', False):
    wd = os.path.dirname(sys.executable)
else:
    wd = ''

MESSAGE_JSON_PATH = os.path.join(wd,"messages.json")
MEDIA_PATH = os.path.join(wd,'media')
EXPORTED_PATH = os.path.join(wd,'exported')

if not os.path.isfile(MESSAGE_JSON_PATH):
    print("No messages.json file found. Please put executable file in the same directory as message.json file.")
    sys.exit()

# Opening JSON file
f = open(MESSAGE_JSON_PATH)

# Load json to a data dictionary
data = json.load(f)

# Close file
f.close()

exclude_partners = ['calllogs', 'Skype']

num_file_found = 0

def utc_to_local(utc_dt):
    return utc_dt.replace(tzinfo=timezone.utc).astimezone(tz=None)

def md5(fname):
    hash_md5 = hashlib.md5()
    with open(fname, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash_md5.update(chunk)
    return hash_md5.hexdigest()

def save_file(src, dst, partner, modTime):
    global num_file_found
    if not os.path.isdir(EXPORTED_PATH):
        os.mkdir(EXPORTED_PATH)
    if not os.path.isdir(os.path.join(EXPORTED_PATH, partner)):
        os.mkdir(os.path.join(EXPORTED_PATH, partner))
    file_to_save = os.path.join(EXPORTED_PATH, partner) + "/" + dst
    if os.path.isfile(file_to_save):
        if md5(src) != md5(file_to_save):
            v = 2
            while os.path.isfile(file_to_save):
                file_parts = os.path.splitext(dst)
                ext = file_parts[1]
                file_name = file_parts[0]
                if md5(src) == md5(file_to_save):
                    file_to_save = ''
                    print("File {} already exist and is identical, skipping...".format(dst + "_v" + str(v-1)))
                    break
                file_parts = os.path.splitext(dst)
                ext = file_parts[1]
                file_name = file_parts[0]
                file_to_save = os.path.join(EXPORTED_PATH, partner) + "/" + file_name + "_v" + str(v) + ext
                v += 1
            print("File {} already exist, writing new version {}".format(dst, v-1))
        else:
            file_to_save = ''
            print("File {} already exist and is identical, skipping...".format(dst))
    if file_to_save != '':
        shutil.copy2(src, file_to_save)
        os.utime(file_to_save, (modTime, modTime))
        num_file_found += 1


# Iterating through the conversations list
for conversation in data['conversations']:
    partner = conversation['id'].split(':')[1]
    if partner.endswith('cast.skype'):
        continue
    if conversation['displayName'] != None:
        partner = conversation['displayName']

    if partner not in exclude_partners and len(conversation['MessageList']) > 0:

        # Iterating through message list, only look for media files
        for message in conversation['MessageList']:

            if message['content'] != '':
                if message['displayName'] != None:
                    partner = message['displayName']

                timestamp = dateutil.parser.isoparse(message['originalarrivaltime'])
                timestamp = utc_to_local(timestamp)
                filename_ts = timestamp.strftime("%Y%m%d-%H%M%S%f-%z")
                modTime = time.mktime(timestamp.timetuple())

                if message['messagetype'] == 'RichText/Media_Video':

                    if message['content'].startswith("<URIObject"):
                        obj_tree = ET.ElementTree(ET.fromstring(message['content']))
                        root = obj_tree.getroot()
                        orig_name_tag = root.find("OriginalName")
                        filename_to_save = ""
                        if orig_name_tag != None and "v" in orig_name_tag.attrib and orig_name_tag.attrib['v'] != "":
                            filename_to_save = orig_name_tag.attrib['v']
                        doc_id = root.attrib["uri"].split("/")[-1]
                        
                        # Find file in media folder
                        if os.path.isfile(MEDIA_PATH + "/" + doc_id + ".1"):
                            file_to_copy = MEDIA_PATH + "/" + doc_id + ".1"
                            if filename_to_save == "":
                                kind = filetype.guess(file_to_copy)
                                if not kind is None:
                                    filename_to_save = doc_id + "." + kind.extension

                            target_file = filename_to_save if filename_to_save != "" else doc_id + ".mp4"
                            save_file(file_to_copy, target_file, partner, modTime)
                        
                elif message['messagetype'] == 'RichText/Media_AudioMsg':

                    if message['content'].startswith("<URIObject"):
                        obj_tree = ET.ElementTree(ET.fromstring(message['content']))
                        root = obj_tree.getroot()
                        orig_name_tag = root.find("OriginalName")
                        filename_to_save = ""
                        if orig_name_tag != None and "v" in orig_name_tag.attrib and orig_name_tag.attrib['v'] != "":
                            filename_to_save = orig_name_tag.attrib['v']
                        doc_id = root.attrib["uri"].split("/")[-1]
                        
                        # Find file in media folder
                        if os.path.isfile(MEDIA_PATH + "/" + doc_id + ".1"):
                            file_to_copy = MEDIA_PATH + "/" + doc_id + ".1"
                            if filename_to_save == "":
                                kind = filetype.guess(file_to_copy)
                                if not kind is None:
                                    filename_to_save = doc_id + "." + kind.extension

                            target_file = filename_to_save if filename_to_save != "" else doc_id + ".m4a"
                            save_file(file_to_copy, target_file, partner, modTime)

                elif message['messagetype'] == 'RichText/Media_GenericFile':

                    if message['content'].startswith("<URIObject"):
                        obj_tree = ET.ElementTree(ET.fromstring(message['content']))
                        root = obj_tree.getroot()
                        orig_name_tag = root.find("OriginalName")
                        filename_to_save = ""
                        extension = ""
                        if orig_name_tag != None and "v" in orig_name_tag.attrib and orig_name_tag.attrib['v'] != "":
                            filename_to_save = orig_name_tag.attrib['v']
                            extension = filename_to_save.split(".")[-1]
                        doc_id = root.attrib["uri"].split("/")[-1]

                        file_found = False
                        
                        # Find file in media folder
                        if os.path.isfile(MEDIA_PATH + "/" + doc_id + ".1"):
                            file_to_copy = MEDIA_PATH + "/" + doc_id + ".1"
                            file_found = True
                        elif os.path.isfile(MEDIA_PATH + "/" + doc_id + "." + extension):
                            file_to_copy = MEDIA_PATH + "/" + doc_id + "." + extension
                            file_found = True

                        if file_found:
                            if filename_to_save == "":
                                kind = filetype.guess(file_to_copy)
                                if not kind is None:
                                    filename_to_save = doc_id + "." + kind.extension

                            target_file = filename_to_save if filename_to_save != "" else doc_id
                            save_file(file_to_copy, target_file, partner, modTime)

                elif message['messagetype'] == 'RichText/UriObject':

                    if message['content'].startswith("<URIObject"):
                        uri_obj_string = message['content'].split("<c_i")[0]
                        obj_tree = ET.ElementTree(ET.fromstring(uri_obj_string))
                        root = obj_tree.getroot()

                        orig_name_tag = root.find("OriginalName")
                        filename_to_save = ""
                        extension = ""
                        if orig_name_tag != None and "v" in orig_name_tag.attrib and orig_name_tag.attrib['v'] != "":
                            filename_to_save = orig_name_tag.attrib['v']
                            extension = filename_to_save.split(".")[-1]
                        doc_id = root.attrib["uri"].split("/")[-1]

                        file_found = False
                        
                        # Find file in media folder
                        if os.path.isfile(MEDIA_PATH + "/" + doc_id + ".1." + extension):
                            file_to_copy = MEDIA_PATH + "/" + doc_id + ".1." + extension
                            file_found = True
                        elif os.path.isfile(MEDIA_PATH + "/" + doc_id + ".1"):
                            file_to_copy = MEDIA_PATH + "/" + doc_id + ".1"
                            kind = filetype.guess(file_to_copy)
                            if not kind is None:
                                filename_to_save = doc_id + "." + kind.extension

                        if file_found:
                            target_file = filename_to_save if filename_to_save != "" else doc_id
                            save_file(file_to_copy, target_file, partner, modTime)

                elif message['messagetype'] == 'RichText/Files':

                    if message['content'].startswith("<files"):
                        files_string = message['content'].split("<e_m")[0]
                        files_tree = ET.ElementTree(ET.fromstring(files_string))
                        root = files_tree.getroot()

                        file_tag = root.find("file")
                        filename_to_save = ""
                        if file_tag != None and file_tag.text != '':
                            filename_to_save = file_tag.text
                            
                            # Find file in media folder
                            if os.path.isfile(MEDIA_PATH + "/" + filename_to_save):
                                file_to_copy = MEDIA_PATH + "/" + filename_to_save
                                save_file(file_to_copy, filename_to_save, partner, modTime)

                else:
                    pass

print("{} files found.".format(num_file_found))